"use strict";

/*
 * polynode-boilerplate-webserver
 *
 * Released under MIT license. Copyright 2019 Jorge Duarte Rodriguez <info@malagadev.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this
 * software and associated documentation files (the "Software"), to deal in the Software
 * without restriction, including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies
 * or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * $Id:$
 *
 * 
 * @format
 *
 */
const Express = require('express');

const BodyParser = require('body-parser');

const http = require('http');

const cors = require('cors');

const RequestContext = require('./RequestContext');

const {
  NotFoundError
} = require('./Errors');

const ROUTE_METHODS = ['get', 'post', 'patch', 'delete', 'put', 'all'];
const ROUTE_METHODS_CORS_NON_COMPLEX = ['get', 'head', 'post'];
const Constants = {
  DEFAULT_SUCCESS_HTTP_STATUS_CODE: 200,
  DEFAULT_ERROR_HTTP_STATUS_CODE: 500
};
const CORS_OPTIONS = {
  origin: (origin, callback) => {
    const allowedByCors = true;

    if (allowedByCors) {
      callback(null, true);
    } else {
      callback(new Error(`${origin} not allowed by CORS`));
    }
  },
  exposedHeaders: ['X-Auth-Refresh-In'] // @todo : move this logic to Configuration (project-dependent)

};

const enableCors = () => cors(CORS_OPTIONS);

const addRequestContext = ({
  getServerHandler,
  getRouteOptions,
  log
}) => {
  // @TODO: definir comportamientos para error o para resolver (extraer toda esa logica de context).
  return (req, res, next) => {
    // log.trace({}, 'Request starts addRequestContext');
    req._context = new RequestContext({
      getRequest: () => req,
      getResponse: () => res,
      next: (...args) => next(...args),
      getServerHandler,
      routeOptions: getRouteOptions(),
      log
    }); // req._context.log({}, 'Request Context has been created.');

    req._context.next();
  };
};

const resolveWithError = (err, responseContentType, res, log) => {
  const httpStatusCode = err.httpStatusCode || Constants.DEFAULT_ERROR_HTTP_STATUS_CODE;
  const outputObj = 'expose' in err && err.expose != null && err.expose || {};
  return sendResponse({
    httpStatusCode,
    outputObj,
    responseContentType,
    log,
    res
  });
};

const getErrorHandler = ({
  log
}) => {
  return (err, req, res, next) => {
    req._context.log({
      err
    }, 'Request error: ' + err.message, 'debug');

    resolveWithError(err, req._context.getServerHandler().getConfig().defaultOutputContentType, res, log);
  };
};

const getContextualisedMiddleware = handler => (req, res, next) => {
  handler({
    _unsafe: req.params
  }, {
    _unsafe: req.body
  }, req._context);
};

const getContextualizedRouteMiddlewares = routeMiddlewares => routeMiddlewares.map(mw => getContextualisedMiddleware(mw));

const getNonFoundHandler = ({
  log
}) => {
  return (req, res, next) => {
    resolveWithError(new NotFoundError(), req._context.getServerHandler().getConfig().defaultOutputContentType, res, log);
  };
};

const getEnhancedRouteMiddlewarePipe = ({
  serverRequestHooks,
  getServerHandler,
  mainMiddleware,
  getRouteOptions,
  log
}) => [addRequestContext({
  getServerHandler,
  getRouteOptions,
  log
}), ...(serverRequestHooks.pre ? getContextualizedRouteMiddlewares(serverRequestHooks.pre) : []), getContextualisedMiddleware(mainMiddleware), getNonFoundHandler(getServerHandler().getDepsContainer()), getErrorHandler(getServerHandler().getDepsContainer())];

const corsMiddleware = (query, body, context) => {
  return new Promise((resolve, reject) => {
    enableCors()(context.getRequest(), context.getResponse(), () => {
      resolve(true);
    });
  });
};

const getEnhancedRouteInstance = (methodName, uri, specificRouteMiddlewares, enhancedRouteHandlers, {
  expressRouter,
  serverRequestHooks,
  getServerHandler,
  log
}) => {
  // console.log('enhancedRouteHandlers: ', enhancedRouteHandlers);
  const obj = {
    enhancedRouteHandlers,
    compiledEnhancedRouteHandlers: {},
    applyEnhancedRouteHandlers: function () {
      enhancedRouteHandlers.forEach(erh => {
        //  console.log('this is: ', this);
        // console.log('process ERH: ', { [erh.name]: erh.handler });
        this[erh.name] = (...args) => {
          //      console.log('[' + erh.name + '] handler is: ', erh.handler, 'args:', args);
          try {
            const {
              handler
            } = erh; // console.log('- this is: ', this);
            // console.log('handler*: ', handler);
            //    console.log('replace this[' + erh.name + ']');

            this.compiledEnhancedRouteHandlers[erh.name] = handler.bind(this)(...args); //      console.log('OK: ');
          } catch (err) {
            console.log('ERH ERROR: ', err);
            return;
          } //  console.log('this is: ', this);


          return this;
        };
      }, {});
    },
    methodName,
    uri,
    expressRouter,
    serverRequestHooks,
    getServerHandler,
    _options: {},
    addRealExpressRoute: function (middlewares) {
      //    console.log('add real express route(X): ', middlewares);
      const binderFunc = this.expressRouter[this.methodName].bind(this.expressRouter);
      const middlewaresToBind = isComplexCorsMethod(this.methodName) ? [enableCors(), ...middlewares] : middlewares;
      const mainResult = binderFunc(this.uri, ...middlewaresToBind);
      return mainResult;
    },
    setOptions: function (opts) {
      this._options = { ...this._options,
        ...opts
      };
      return this;
    },
    getOptions: function () {
      return this._options;
    },
    resolve: async function () {
      const finalNext = async (query, body, context, transform) => {
        //    console.log('specificRouteMiddlewares: ', specificRouteMiddlewares);
        // @todo : en el caso de que "cur" no sea una función (o mas bien se haga una excepcion)
        // alertar de que hay una ruta mal configurada (apuntando hacia algo que no es una funcion)
        //  console.log('finalnext: ', query, body, transform);
        //  ;
        const result = await specificRouteMiddlewares.reduce((last, cur) => last.then(() => cur(query, body, context)), Promise.resolve());

        if (transform) {
          return transform(result);
        } //    console.log('Result: ', result);


        return result;
      };

      const middlewares = getEnhancedRouteMiddlewarePipe({
        serverRequestHooks,
        getServerHandler,
        log,
        mainMiddleware: async (query, body, context) => {
          const disableAutoFlush = 'disableAutoFlush' in this.getOptions() && this.getOptions().disableAutoFlush === true; //  console.log('enhancedRouteHandlers.reduce starts:', [query, body, context]);

          try {
            const _x = await Object.keys(this.compiledEnhancedRouteHandlers).reduce((last, cur, idx) => {
              return last.then(lastRes => {
                //      console.log('Last result was: ', lastRes);
                const params = lastRes || [query, body, context, null];
                /*        console.log(
                  'Call handler "' +
                    cur +
                    '" (last: ' +
                    (idx - 1 in enhancedRouteHandlers && enhancedRouteHandlers[idx - 1].name) +
                    ') params: ',
                 params
                ); */
                // console.log('New params are: ', params);

                /*          console.log(
                  'going to exec: ',
                  cur,
                  'params: ',
                  params,
                  'context currentUser:',
                  params[2].currentUser
                );
                */

                const result = this.compiledEnhancedRouteHandlers[cur].bind(this)(...params); // console.log('reuslt is : ', { result });

                return result;
              });
            }, Promise.resolve()); //            console.log('_x:', _x);


            const _r = await _x; //          console.log('_r: ', _r);


            const result = await finalNext(...(_r || [query, body, context, null])); //  console.log('r2: ', result);

            if (!disableAutoFlush) {
              context.resolve(result);
            }
          } catch (err) {
            return context.reject(err);
          }
        },
        getRouteOptions: () => this.getOptions()
      });
      /*
      */

      return this.addRealExpressRoute(middlewares);
    }
  }; // console.log('obj is: ', obj);

  obj.applyEnhancedRouteHandlers();
  return obj;
};

const isComplexCorsMethod = methodName => methodName in ROUTE_METHODS_CORS_NON_COMPLEX === false;

const getBypassedRouterMethods = ({
  expressRouter,
  getServerHandler,
  serverRequestHooks,
  enhancedRouteHandlers,
  log
}) => {
  return ROUTE_METHODS.reduce((res, methodName) => {
    res[methodName] = (uri, ...specificRouteMiddlewares) => {
      return getEnhancedRouteInstance(methodName, uri, isComplexCorsMethod(methodName) ? [corsMiddleware, ...specificRouteMiddlewares] : specificRouteMiddlewares, enhancedRouteHandlers, {
        getServerHandler,
        log,
        serverRequestHooks,
        expressRouter
      });
    };

    return res;
  }, {});
};

const configureExpressApp = app => {
  app.use(BodyParser.json({
    limit: '50mb'
  }));
  app.use(enableCors());
  /*
  app.use(function(req, res, next) {
    res.header('Access-Control-Allow-Origin', 'camperalia.com'); // update to match the domain you will make the request from
    res.header('Access-Control-Allow-Headers', 'Origin, X-Requested-With, Content-Type, Accept');
    next();
  });
  */
};

const createRouter = ({
  getServerHandler,
  serverRequestHooks,
  getEnhancedRouteHandlers,
  log
}) => {
  const expressRouter = Express.Router();
  const router = getBypassedRouterMethods({
    expressRouter,
    getServerHandler,
    serverRequestHooks,
    enhancedRouteHandlers: getEnhancedRouteHandlers(),
    log
  });
  return {
    router,
    expressRouter
  };
};

const configureExpressRouting = (app, {
  expressRouter
}) => {
  app.use(expressRouter);
};

const serverListen = (server, port) => new Promise((resolve, reject) => {
  server.listen(port, data => {
    resolve(data);
  });
});

const sendResponse = ({
  res,
  log,
  httpStatusCode,
  responseContentType,
  outputObj
}) => {
  res.setHeader('Content-Type', responseContentType);
  res.status(httpStatusCode).json(outputObj).end();
};

module.exports = depsContainer => {
  const log = depsContainer.log;
  log.info({}, '[boilerplate-webserver] ##### Initializing... ');
  const config = depsContainer.config;
  log.debug({}, 'web server config is: ', depsContainer.config, '.enhanceRequestContext: ', depsContainer.enhanceRequestContext);
  const expressHandler = Express();

  const getServerHandlerForContext = () => ({
    getConfig: () => config,
    getDepsContainer: () => depsContainer,
    sendResponse
  });

  const enhancedRouteHandlers = [];
  const mainRouter = createRouter({
    getServerHandler: getServerHandlerForContext,
    serverRequestHooks: depsContainer.serverRequestHooks,
    getEnhancedRouteHandlers: () => enhancedRouteHandlers,
    log
  });
  configureExpressApp(expressHandler);
  configureExpressRouting(expressHandler, mainRouter);
  const server = http.createServer(expressHandler);
  const instance = {
    getMainRouter: () => mainRouter.router,
    getServer: () => server,
    startServer: async () => {
      await serverListen(server, config.port);
      log.debug({}, '*** Server listening in port: ' + config.port);
      return true;
    },
    sendResponse,
    getConfig: () => config,
    getDepsContainer: () => depsContainer,
    registerEnhancedRouteHandlers: newRouteHandlers => {
      const entries = Object.entries(newRouteHandlers);

      for (const [name, handler] of entries) {
        enhancedRouteHandlers.push({
          name,
          handler
        });
      }
    }
  };

  if (depsContainer.enhanceServerInstance) {
    // @todo: we need to find a way to affect getServerHandlerForContext() behavior efficiently
    depsContainer.enhanceServerInstance.call(instance);
  }

  return instance;
};